
(* This file is part of Numerix.  Numerix is free software; you can
   redistribute it and/or modify it under the terms of the GNU General
   Public License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA *)

(* +------------------------------------------------------------------------+
   |                                                                        |
   |              Implmentations diverses des entiers relatifs             |
   |              Various implementations of signed integers                |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

(* M. Quercia, 11/02/2001 *)

(* Rmq : pour faire plaisir au prprocesseur, chaque ligne doit contenir   *)
(* un nombre PAIR d''appostrophes, y compris celles en commentaire...      *)
(* Caution : the preprocessor (cpp) requires that an EVEN number of quotes *)
(* resides on every line, including those inside of Ocaml comments...      *)

module type Int_type = sig

  type t        (* reprsentation d''un entier, abstract integer *)
  type tref     (* entier mutable,              mutable integer  *)

  val name : unit -> string   (* nom du module, name of module *)

  (* oprations unaires et binaires, unary and binary operations *)
  (*  op           r   s  a    b               effet,effect               *)
  val add       :         t -> t -> t      (*  a + b                      *)
  val sub       :         t -> t -> t      (*  a - b                      *)
  val mul       :         t -> t -> t      (*  a * b                      *)
  val quo       :         t -> t -> t      (*  a div b                    *)
  val modulo    :         t -> t -> t      (*  a mod b                    *)
  val quomod    :         t -> t -> t*t    (*  (a div b, a mod b)         *)
  val abs       :         t ->      t      (*  |a|                        *)
  val neg       :         t ->      t      (*  -a                         *)
  val sqr       :         t ->      t      (*  a^2                        *)
  val sqrt      :         t ->      t      (*  trunc(a^(1/2))             *)
  val powmod    :     t-> t -> t -> t      (*  s^a mod b, a >= 0          *)
                                                                           
  val add_in    : tref -> t -> t -> unit   (*  r := a + b                 *)
  val sub_in    : tref -> t -> t -> unit   (*  r := a - b                 *)
  val mul_in    : tref -> t -> t -> unit   (*  r := a * b                 *)
  val quo_in    : tref -> t -> t -> unit   (*  r := a div b               *)
  val quomod_in : tref->tref->t->t ->unit  (*  r := a div b, s := a mod b *)
  val abs_in    : tref -> t ->      unit   (*  r := |a|                   *)
  val neg_in    : tref -> t ->      unit   (*  r := -a                    *)
  val sqr_in    : tref -> t ->      unit   (*  r := a^2                   *)
  val sqrt_in   : tref -> t ->      unit   (*  r := sqrt(a)               *)
  val powmod_in : tref-> t->t->t -> unit   (*  r := s^a mod b, b>=0       *)

  (* gestion des rfrences, mutable integers handling *)
  val make_ref  :         t -> tref        (*  ref(copy a)                *)
  val copy_in   : tref -> t ->      unit   (*  r := a                     *)
  val copy_out  : tref ->           t      (*  copy(!r)                   *)
  val look      : tref ->           t      (*  !r                         *)

  (* oprations mixtes, mixed operations *)
  (*  op           r   s  a      b             effet,effect               *)
  val add_1     :         t -> int -> t    (*  a + b                      *)
  val sub_1     :         t -> int -> t    (*  a - b                      *)
  val mul_1     :         t -> int -> t    (*  a * b                      *)
  val quo_1     :         t -> int -> t    (*  a div b                    *)
  val mod_1     :         t -> int -> int  (*  a mod b                    *)
  val quomod_1  :         t -> int -> t*int(*  (a div b, a mod b)         *)
  val pow       :         t -> int -> t    (*  a ^ b                      *)
  val root      :         t -> int -> t    (*  trunc(a ^ 1/b)             *)
  val fact      :       int ->        t    (*  a!                         *)
  val shr       :         t -> int -> t    (*  sgn(a)*(|a| div 2^b)       *)
  val shl       :         t -> int -> t    (*  a * 2^b                    *)
  val split     :         t -> int -> t*t  (*  sgn(a)*(|a| div 2^b, |a| mod 2^b) *)
  val join      :    t -> t -> int -> t    (*  s + a*2^b                  *)
                                                                          
  val add_1_in  : tref -> t -> int -> unit (*  r := a + b                 *)
  val sub_1_in  : tref -> t -> int -> unit (*  r := a - b                 *)
  val mul_1_in  : tref -> t -> int -> unit (*  r := a * b                 *)
  val quomod_1_in : tref->t -> int -> int  (*  a mod b, r := a div b      *)
  val pow_in    : tref -> t -> int -> unit (*  r := a ^ b                 *)
  val root_in   : tref -> t -> int -> unit (*  r := trunc(a ^ 1/b)        *)
  val fact_in   : tref ->      int -> unit (*  r := a!                    *)
  val shl_in    : tref -> t -> int -> unit (*  r := a shl b               *)
  val shr_in    : tref -> t -> int -> unit (*  r := a shr b               *)
  val split_in  : tref->tref->t->int->unit (* (r,s) := a split b          *)
  val join_in   : tref->t->t -> int ->unit (* r := s + a*2^b              *)

  (* pgcd et fonctions associes
    ----------------------------

    [cfrac a b] retourne des entiers p,q,u,v,d tels que :

       u*a - v*b = d = pgcd positif(a,b)
       p*b = q*a = ppcm(a,b) avec le signe de a*b
       p*u - q*v = 1

    [gcd a b] retourne seulement d et [gcd_ex a b] retourne u,v,d

    gcd and associated functions
    ----------------------------

    [cfrac a b] returns integers p,q,u,v,d such that :

       u*a - v*b = d = positive gcd of a and b
       p*b = q*a = lcm(a,b) with the same sign as a*b
       p*u - q*v = 1

    [gcd a b] returns only d and [gcd_ex a b] returns u,v,d
  *)

  val gcd       :         t -> t -> t
  val gcd_ex    :         t -> t -> t*t*t
  val cfrac     :         t -> t -> t*t*t*t*t

  val gcd_in    : tref -> t -> t -> unit
  val gcd_ex_in : tref->tref->tref->t->t-> unit
  val cfrac_in  : tref->tref->tref->tref->tref->t->t-> unit

  (* comparaison, comparison *)                                                      
  (*  op           r      a    b               effet,effect               *)
  val eq        :         t -> t -> bool   (*  a = b ?                    *)
  val neq       :         t -> t -> bool   (*  a <> b ?                   *)
  val inf       :         t -> t -> bool   (*  a < b ?                    *)
  val infeq     :         t -> t -> bool   (*  a <= b ?                   *)
  val sup       :         t -> t -> bool   (*  a > b ?                    *)
  val supeq     :         t -> t -> bool   (*  a >= b ?                   *)
  val cmp       :         t -> t -> int    (*  signe(a-b) (1,-1,0)        *)
  val eq_1      :         t ->int-> bool   (*  a = b ?                    *)
  val neq_1     :         t ->int-> bool   (*  a <> b ?                   *)
  val inf_1     :         t ->int-> bool   (*  a < b ?                    *)
  val infeq_1   :         t ->int-> bool   (*  a <= b ?                   *)
  val sup_1     :         t ->int-> bool   (*  a > b ?                    *)
  val supeq_1   :         t ->int-> bool   (*  a >= b ?                   *)
  val cmp_1     :         t ->int-> int    (*  signe(a-b) (1,-1,0)        *)
  val sgn       :         t ->      int    (*  signe(a)   (1,-1,0)        *)
  val max       :         t -> t -> t      (*  max(a,b)                   *)
  val min       :         t -> t -> t      (*  min(a,b)                   *)

                                                                        
  (* oprations diverses, miscellaneous operations *)                                              
  (*  op           r      a    b               effet,effect               *)
  val nbits     :         t ->        int  (*  taille en bits de |a|, size in bits of |a| *)
  val lowbits   :         t ->        int  (*  bits 0..30 de |a|          *)
  val highbits  :         t ->        int  (*  bits n-31..n-1, n=nbits(a) *)
  val nth_bit   :         t -> int -> bool (*  bit  b de |a|              *)
  val nth_word  :         t -> int -> int  (*  bits 16b..16b+15 de |a|    *)
  val random_bits :       int      -> t    (*  random number between 0 and 2^a - 1 *)

  val of_int    : 	  int    ->   t   
  val int_of    : 	  t      ->   int   
  val string_of : 	  t      ->   string
  val of_string : 	  string ->   t
  val toplevel_print :    t      ->   unit

end

(* notation infixe, infix notation *)
module Infixes(E : Int_type) : sig
  open E

  val (  +  )  :   t -> t   -> t      (* add      *)
  val (  -  )  :   t -> t   -> t      (* sub      *)
  val (  *  )  :   t -> t   -> t      (* mul      *)
  val (  /  )  :   t -> t   -> t      (* quo      *)
  val (  %  )  :   t -> t   -> t      (* modulo   *)
  val ( /%  )  :   t -> t   -> t*t    (* quomod   *)
  val ( <<  )  :   t -> int -> t      (* shl      *)
  val ( >>  )  :   t -> int -> t      (* shr      *)
  val (  ** )  :   t -> int -> t      (* pow      *)
            	   
  val (  +. )  :   t -> int -> t      (* add_1    *)
  val (  -. )  :   t -> int -> t      (* sub_1    *)
  val (  *. )  :   t -> int -> t      (* mul_1    *)
  val (  /. )  :   t -> int -> t      (* quo_1    *)
  val (  %. )  :   t -> int -> int    (* mod_1    *)
  val ( /%. )  :   t -> int -> t*int  (* quomod_1 *)
            
  val (  +! )  : int -> int -> int    (* add      *)
  val (  -! )  : int -> int -> int    (* sub      *)
  val (  *! )  : int -> int -> int    (* mul      *)
  val (  /! )  : int -> int -> int    (* quo      *)
  val (  %! )  : int -> int -> int    (* mod      *)
  val (  =. )  :   t -> int -> bool   (* eq_1     *)
  val ( <>. )  :   t -> int -> bool   (* neq_1    *)
  val (  <. )  :   t -> int -> bool   (* inf_1    *)
  val ( <=. )  :   t -> int -> bool   (* infeq_1  *)
  val (  >. )  :   t -> int -> bool   (* sup_1    *)
  val ( >=. )  :   t -> int -> bool   (* supeq_1  *)
            
  val (  =! )  : 'a  -> 'a  -> bool   (* eq       *)
  val (	<>! )  : 'a  -> 'a  -> bool   (* neq      *)
  val (	 <! )  : 'a  -> 'a  -> bool   (* inf      *)
  val (	<=! )  : 'a  -> 'a  -> bool   (* infeq    *)
  val (	 >! )  : 'a  -> 'a  -> bool   (* sup      *)
  val (	>=! )  : 'a  -> 'a  -> bool   (* supeq    *)

  val ( ~~ )   : tref -> t            (* look     *)
end
module Big   : Int_type        (* bigints de Ocaml *)
module Clong : Int_type        (* longints C *)
module Dlong : Int_type        (* longints C double prcision *)
module Gmp   : Int_type        (* bigints de Gmp *)
(* affiche en secondes le temps du processus, la diffrence avec le temps *)
(* prcdent et la chaine passe en argument.                             *)
(* displays in seconds the CPU time of the process, the difference with   *)
(* the previous display and the given string.                             *)
external chrono : string -> unit = "chrono"

(* Slection  l'excution d'une implmentation des grands entiers :
   Dfinir un foncteur

	Main : Int_type -> Main_type

   implmentant la fonction principale

   	main : string list -> unit

   puis appeler la fonction

	Start(Main).start()

   Cette fonction analyse la ligne de commande, extrait une ventuelle
   option "-e <impl>" spcifiant quelle implmentation des entiers utiliser,
   puis appelle la fonction Main(<impl>).main avec en argument la liste des
   autres paramtres figurant sur la ligne de commande.

                              ----------

   Run-time selection of a specific big integer implementation :
   Define a functor

	Main : Int_type -> Main_type

   providing the main function

   	main : string list -> unit

   and call the function

	Start(Main).start()

   this function will parse the command line, extract an eventual option
   "-e <impl>" telling which big integer module to use, then call the
   function Main(<impl>).main passing the list of remaining command line
   arguments.
*)

module type Main_type = sig
  val main : string list -> unit
end
module Start(Main : functor(E:Int_type) -> Main_type) : sig
  val start : unit -> unit
end
